unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms,
  Dialogs, ExtCtrls, Model3D, Graphics, StdCtrls, ModelBuild,
  ColorConv, World3D, Math, Global;

type

  TForm1 = class(TForm)
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure SetMousePos(APoint : TPoint);
    procedure FormActivate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SetFlatFaces(Flat : boolean);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BlitBitmap;
  private
    { Private declarations }
    World : T3DWorld; // the 3D world
    t : integer; // the time ball has been in air
    MousePos : TPoint; // position that cursor is reset to - center of screen
    MoveForward, MoveBack, MoveRight, MoveLeft, MoveUp, MoveDown : boolean; // flags for moving viewing position
    StarBmp, TextBmp : TBitmap;
    StarPos : TPointFloat; // center position in star background
    FlatFaces : boolean;
    CaptionStart : String;
  public
    { Public declarations }
  end;

procedure DrawWrap(ACanvas1 : TCanvas ; ACanvas2 : TCanvas ; W, H, x, y : integer);

const
  CamSpeed = 60;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
Const
  clDullBlue = $00FCDDDD;
  cldullGreen = $00F5FCF5;
  clDullRed = $00DDDDFC;
  ClubCoordsData : array[0..7] of Real = (0, 16, 32, 360, 568, 616, 780, 792);
  ClubRadiusData : array[0..7] of Real = (26.4, 26.4, 13.6, 24, 64, 64, 33.6, 30);
  ClubColorsData : array[0..8] of TColor = (clwhite, clwhite, clDullBlue, clDullRed, clDullRed, clDullRed, clwhite, clwhite, clwhite);
Var
  ClubCoords, ClubRadius : TFloatArray;
  ClubColors : TColorArray;
  Lx, Ly : integer;
  LHeights : T2DFloatArray;
  LNumFaces : integer;
  LIntensity : byte;
  LRedHLS : THLS;
begin
  Randomize;

  // create the starfield bitmap - background
  StarBmp := TBitmap.Create;
  with StarBmp do begin
    PixelFormat := pf32bit;
    Width := ClientWidth;
    Height := ClientHeight;
    Canvas.Brush.Color := clBlack;
    Canvas.FillRect(Rect(0,0,Width,Height));
    for Lx := 0 to 1000 do begin
      LIntensity := Random(220) + 35;
      Canvas.Pixels[Random(Width), Random(Height)] :=
        (LIntensity shl 16) + (LIntensity shl 8) + LIntensity;
    end;
  end;
  StarPos.X := 0; StarPos.Y := 0;

  // create the text bitmap to display instructions
  TextBmp := TBitmap.Create;
  with TextBmp do begin
    PixelFormat := pf32bit;
    Width := 380;
    Height := 15;
    Canvas.Brush.Color := clBlack;
    Canvas.FillRect(Rect(0,0,380,15));
    Canvas.Font.Color := clWhite;
    Canvas.Font.Name := 'New Courier';
    Canvas.Font.Height := 12;
    Canvas.TextOut(0, 0, 'arrows & mouse buttons - move,  p - phong, g - gouraud, f - flat,  w - wireframe, esc - exit');
    TransparentColor := clBlack;
    Transparent := True;
    PixelFormat := pf4bit;
  end;

  // create dynamic arrays of club data
  SetLength(ClubCoords, 8);
  SetLength(ClubRadius, 8);
  SetLength(ClubColors, 9);
  for Lx := 0 to 7 do begin
    ClubCoords[Lx] := ClubCoordsData[Lx];
    ClubRadius[Lx] := ClubRadiusData[Lx];
    ClubColors[Lx] := ClubColorsData[Lx];
  end;
  ClubColors[8] := clwhite;

  // create an instance of T3DWorld - the main class
  World := T3DWorld.Create;
  with World do begin
    ScreenRect := Rect(0, 0, ClientWidth, ClientHeight);

    // lights
    Light.Position := Point3D(400, -800, 0);
    Light.Luminance := 140;
    Ambient := 75;

    // camera
    Camera.Position := Point3D(0, -60, -400);
    Camera.NearPlaneOffset := 100;
    Camera.FarPlaneOffset := 10000;
    // amount of perspective - distance of eye to screen
    Camera.PerspectiveOffset := 600;
    Camera.ComputeFrustum;

    // action!
    SetLength(Objects, 7);

    // create an instance of a T3DModel - a juggling club
    Objects[0] := T3DModel.Create;
    BuildSolidOfRevolution(Objects[0], ClubCoords, ClubRadius, ClubColors, 20);
    with Objects[0] do begin
      Center := Point3D(0, 440, 0);
      Translate(Point3D(-200, -540, 1400)); // origin is at center of view space
      Rotate(pi, AxisX);
      Color := ColorToRGB(clWhite);
    end;

    // create second object - a ball
    Objects[1] := T3DModel.Create;
    //BuildTube(Objects[1], 100, 60, 3, clDullBlue);
    BuildGeoSphere(Objects[1], icosahedron, 2, 60, clDullBlue, True, True);
    Objects[1].Translate(Point3D(0, 0, 1400));
    Objects[1].Color := ColorToRGB(clBlue);

    // create third object - a juggling club
    Objects[2] := T3DModel.Create;
    Objects[2].Assign(Objects[0]);
    Objects[2].Translate(Point3D(400, 0, 0));

    // create fourth object - a ball
    Objects[3] := T3DModel.Create;
    Objects[3].Assign(Objects[1]);
    with Objects[3] do begin
      LRedHLS := RGBtoHLS(ColorToRGB(clDullRed));
      for Lx := 0 to High(Faces) do Faces[Lx].HLS := LRedHLS;
      Translate(Point3D(-400, 0, 0));
      Color := ColorToRGB(clRed);
    end;

    // create fifth object - a ball
    Objects[4] := T3DModel.Create;
    Objects[4].Assign(Objects[3]);
    Objects[4].Translate(Point3D(800, 0, 0));

    // create sixth object - floor
    SetLength(LHeights, 5);
    for Lx := 0 to 4 do SetLength(LHeights[Lx], 5);
    for Ly := 0 to 4 do for Lx := 0 to 4 do LHeights[Lx][Ly] := 0;
    Objects[5] := T3DModel.Create;
    BuildSurface(Objects[5], LHeights, 250, 200, 0, $00A58F67);
    with Objects[5] do begin
      Center := Point3D(500, 400, 0);
      Translate(Point3D(-500, -400, 0));
      Rotate(pi*0.5, AxisX);
      Translate(Point3D(0, 350, 1400));
      Color := ColorToRGB($00A58F67);
    end;
    // underside
    Objects[6] := T3DModel.Create;
    Objects[6].Assign(Objects[5]);
    Objects[6].Rotate(pi, AxisX);

  end; // with world

  t := 0;

  LNumFaces := 0;
  for Lx := 0 to High(World.Objects) do
    Inc(LNumFaces, length(World.Objects[Lx].Faces));
  CaptionStart := '3D World Test - ' + IntToStr(LNumFaces) + ' faces';
  Caption := CaptionStart + ' - Gouraud';

  FlatFaces := False;

  ShowCursor(false); // make cursor invisible
end;

// calculate screen center and move cursor to this point
procedure TForm1.FormActivate(Sender: TObject);
begin
  MousePos := Point(ClientWidth shr 1, ClientHeight shr 1);
  SetMousePos(MousePos);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
Var
  LBallHeight, LBallHeight2, LBallHeight3, dx, dy : integer;
  t2, t3 : byte;
  AngleX, AngleY, Lsin, Lcos : extended;
  NewMousePos : TPoint;
begin
  // calculate new ball height
  LBallHeight := Sqr(12 - t) shl 2 - 290;

  t2 := (t + 8) mod 24;
  LBallHeight2 := Sqr(12 - t2) shl 2 - 290;

  t3 := (t + 16) mod 24;
  LBallHeight3 := Sqr(12 - t3) shl 2 - 290;

  Inc(t);
  if t > 23 then t := 0;

  // calculate mouse displacement
  NewMousePos := ScreenToClient(Mouse.CursorPos);
  dx := NewMousePos.X - MousePos.X;
  dy := MousePos.Y - NewMousePos.Y;

  with World do begin

    with Camera do begin

      AngleX := 0; AngleY := 0;

      // look up or down - pitch
      if dy <> 0 then begin
        AngleY := dy * 0.003;
        Rotate(AngleY, AxisX);
      end;

      // look left or right
      if dx <> 0 then begin
        AngleX := dx * 0.003;
        SinCos(AngleX, Lsin, Lcos);
        RotateVector(Ahead, Lsin, Lcos, AxisY);
        RotateVector(Right, Lsin, Lcos, AxisY);
        RotateVector(Down , Lsin, Lcos, AxisY);
      end;

      // move viewing position in forward direction
      if MoveForward then Move( CamSpeed, AxisZ) else
      if MoveBack    then Move(-CamSpeed, AxisZ);
      // move viewing position in right direction
      if MoveRight   then Move( CamSpeed, AxisX) else
      if MoveLeft    then Move(-CamSpeed, AxisX);
      // move viewing position in down direction
      if MoveDown    then Move( CamSpeed, AxisY) else
      if MoveUp      then Move(-CamSpeed, AxisY);

    end; // with camera

    // return cursor to center so that there is an 'infinite mouse'
    SetMousePos(MousePos); 

    // rotate club
    Objects[0].Rotate( pi / 24, AxisX);
    Objects[2].Rotate(-pi / 24, AxisX);

    // move ball to the current position
    Objects[1].Translate( Point3D(0, LBallHeight  - Objects[1].Center.Y, 0) );
    Objects[3].Translate( Point3D(0, LBallHeight2 - Objects[3].Center.Y, 0) );
    Objects[4].Translate( Point3D(0, LBallHeight3 - Objects[4].Center.Y, 0) );

    // calculate new star background position
    if dx <> 0 then begin
      StarPos.X := StarPos.X - AngleX * Camera.PerspectiveOffset;
      if StarPos.X > ScreenRect.Right then
        StarPos.X := StarPos.X - ScreenRect.Right else
      if StarPos.X < 0 then
        StarPos.X := StarPos.X + ScreenRect.Right;
    end;
    if dy <> 0 then begin
      StarPos.Y := StarPos.Y + AngleY * Camera.PerspectiveOffset;
      if StarPos.Y > ScreenRect.Bottom then
        StarPos.Y := StarPos.Y - ScreenRect.Bottom else
      if StarPos.Y < 0 then
        StarPos.Y := StarPos.Y + ScreenRect.Bottom;
    end;

    // clear the off-screen bitmap by drawing the star background in position
    DrawWrap(StarBmp.Canvas, OffScrBmp.Canvas,
             ScreenRect.Right, ScreenRect.Bottom,
             Round(StarPos.X), Round(StarPos.Y));

    // render the world to the off-screen bitmap
    Render;

    World.OffScrBmp.Canvas.Draw(11, 385, TextBmp);

    // display the bitmap on the screen
    BlitBitmap;

  end; // with world
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_UP     : MoveUp := True;
    VK_DOWN   : MoveDown := True;
    VK_RIGHT  : MoveRight := True;
    VK_LEFT   : MoveLeft := True;
    VK_ESCAPE : Close;
    Ord('0')  : World.SuperSample := 0;
    Ord('1')  : World.SuperSample := 2;
    Ord('2')  : World.SuperSample := 4;
    Ord('3')  : World.SuperSample := 6;
    Ord('4')  : World.SuperSample := 8;
    Ord('W')  : begin
                  World.Wireframe := True;
                  Caption := CaptionStart + ' - Wireframe';
                end;
    Ord('P')  : begin
                  World.Phong := True;
                  World.Wireframe := False;
                  SetFlatFaces(False);
                  Caption := CaptionStart + ' - Phong';
                end;
    Ord('G')  : begin
                  World.Phong := False;
                  World.Wireframe := False;
                  SetFlatFaces(False);
                  Caption := CaptionStart + ' - Gouraud';
                end;
    Ord('F')  : begin
                  SetFlatFaces(True);
                  World.Wireframe := False;
                  Caption := CaptionStart + ' - Flat';
                end;
  end;
end;

procedure TForm1.SetFlatFaces(Flat : boolean);
Var
  LObject, LFace : integer;
begin
  if Flat = FlatFaces then Exit;
  for LObject := 0 to High(World.Objects) do
    for LFace := 0 to High(World.Objects[LObject].Faces) do
      if length(World.Objects[LObject].Faces[LFace].Corners) < 5 then
        World.Objects[LObject].Faces[LFace].Curved := FlatFaces;
  FlatFaces := Flat;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_UP    : MoveUp := False;
    VK_DOWN  : MoveDown := False;
    VK_RIGHT : MoveRight := False;
    VK_LEFT  : MoveLeft := False;
  end;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then MoveForward := True
                     else MoveBack := True;
end;

procedure TForm1.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MoveForward := False;
  MoveBack := False;
end;

// draw the off screen bitmap to the forms canvas
procedure TForm1.BlitBitmap;
begin
  BitBlt(Canvas.Handle,
         0,
         0,
         400,
         400,
         World.OffScrBmp.Canvas.Handle,
         0,
         0,
         SrcCopy);
end;

// set the position of the cursor to a specified point
procedure TForm1.SetMousePos(APoint : TPoint);
begin
  APoint := ClientToScreen(APoint);
  SetCursorPos(APoint.X, APoint.Y);
end;

// copy a canvas to a given location on another canvas with wrap-round
// canvas' must have equal dimensions
procedure DrawWrap(ACanvas1, ACanvas2 : TCanvas ; W, H, x, y : integer);
Var
  V, U : integer;          //   __________       ____W_____
  SRect, DRect : TRect;    //  | A |  B   |     |      |   |
begin                      //  |---p------| ==> |  C   | D |H    p = (x,y)
  V := W - x;              //  | D |  C   |     |------p---|
  U := H - y;              //  |___|______|     |__B___|_A_|
                           //    ACanvas1         ACanvas2
  // Copy rect A into position
  SRect := Rect(0, 0, V, U);
  DRect := Rect(x, y, W, H);
  ACanvas2.CopyRect(DRect, ACanvas1, SRect);

  // Copy rect B into position
  SRect.Left := V; SRect.Right := W;
  DRect.Left := 0; DRect.Right := x;
  ACanvas2.CopyRect(DRect, ACanvas1, SRect);

  // Copy rect C into position
  SRect.Top := U; SRect.Bottom := H;
  DRect.Top := 0; DRect.Bottom := y;
  ACanvas2.CopyRect(DRect, ACanvas1, SRect);

  // Copy rect D into position
  SRect.Left := 0; SRect.Right := V;
  DRect.Left := x; DRect.Right := W;
  ACanvas2.CopyRect(DRect, ACanvas1, SRect);
end;

end.
